# $Id: Setpwd.pm 47261M 2018-01-15 12:11:04Z (local) $
# $URL: https://svn.uvt.nl/its-id/trunk/sources/pwdmodifier/keymaster/lib/Setpwd.pm $
use strict;
use warnings  FATAL=> 'all';

package Setpwd;
use Data::Dumper;
use Carp qw (cluck confess);
use Encode qw(encode);
use MIME::Base64;
use Digest::SHA;
use FileHandle;
use Fcntl qw(:DEFAULT :flock);
use IO::Seekable;
use Rcrypt;
use Net::LDAP;
use Net::LDAPS;
use Net::LDAP::Extension::SetPassword;
use POSIX qw(strftime mktime);

use parent qw(Baseobject);
our @EXPORT = qw (verbosity);
my $FAIL = "keymaster status: Failure:";

sub new {
	my $pkg = shift;
	my $self = $pkg->SUPER::new(@_);
	#
	# NOTE NOTE NOTE allow all fields is set to  YES
	#
	$self->{allowallfields} = 1;

	$self->needs(
		[ qw (
		   ipaddress anr ldaphost nosvnfile ldapservers capath oracle_home userinfo
		   oracle_pwd_read_table oracle_pwd_write_table allowpwdrepeat sql passwordLeader dir
		   ) ]
		);

	$self->allows(
		[ qw (challenge time oracle_testanr mailalso2 skiporacle dryrun untouchables newPwd 
              silentsshclients silentclients contact language
	          syncOnlyIfNeeded verifyPasswordSet adminMailAddress ldapRetries disableReason ) ]
		);

	$self->allows({'tildeusername' => ''});
	my $cf = shift;
	$self->{cf} = $cf;
	$self->fields(@_);
	$self->fields($self->getparms($cf));
	$self->checkneeds();
	return $self if $self->error();
	$self->getAdminAccounts();
	return $self if $self->error();

	$self->{fields}->{allowpwdrepeat} = 1 if $self->{fields}->{stealthmode};
	$self->saveLdapObjects();
	$self->checkAccountInSQL() unless $self->{fields}->{skiporacle};
	$self->debug("Setpwd initialised");
	return $self;
}

sub verifyPasswordSet {
  # Verify that the new password can actually be used
	my ($self, $host) = @_;
	my $pwd = $self->{ldapaccounts}->{$host}->{verificationPassword};
		
	my $port = 636; #fixme!
	my $uid = $self->{fields}->{userinfo}->{uid};
	my @hosts = (keys %{$self->{ldapobjects}});
	@hosts = ($host) if $host;

	my $errorString = '';

	foreach my $host (@hosts) {
		my $ldap;
		my $retries = $self->{fields}->{ldapRetries};
		my $attempt = 0;
		my $state;
		while (!$ldap && $attempt++ < $retries) {
			$ldap = Net::LDAPS->new(
				$host,
#      FIXME: verify uitgezet voor AD
#      'verify' => 'require',
				'capath' => $self->{fields}->{capath},
				'port'  => $port,
				);
			$state = $@;
			warn "connect attempt $attempt FAILED for: $host ($uid)" unless $ldap;
		}

		unless ($ldap) {
			warn "ERROR Could not connect to $host:$port ($uid) $state";
			$errorString .= "$uid\@$host: NOCONNECT ";
			next;
		}
		$self->debug("Connection to $host OK");
		foreach my $entry ( @{$self->{ldapobjects}->{$host}->{entries}}){
			my $dn = $entry->dn;

#			next unless $self->isPersonClass($entry, $host);
			my $msg = $ldap->bind($dn, password=>$pwd);
			if ($msg->code)	{
				warn "verification bind FAILED for: $dn ($uid) on ldapserver: $host:$port ",$msg->error();
				$errorString .= "Verification bind failed for: $uid\@$host";
			}else {
				$self->debug("verification bind OK for dn: '$dn' ($uid) on ldapserver: $host:$port");
			}
		}
	}
	$self->error($errorString) if $errorString;
}

sub saveLdapObjects {
	my $self = shift;

	foreach my $host ( keys %{$self->{ldapaccounts}}) {
		$self->debug ("host $host");
		unless (exists ($self->{ldapaccounts}->{$host}->{protocol})) {
			$self->verbose("IGNORING $host");
			next ;
		}
		$self->saveldapobject($host);
	}
}

sub fetchCurrentSHAPassword {
	my $self = shift;
	my $anr = $self->{fields}->{anr};
	my $sql = $self->{fields}->{sql};
	return $sql->getCurrentSHAPassword($anr);
}

sub getPasswordChangeCount {
	my $self = shift;
	my $anr = $self->{fields}->{anr};
	my $sql = $self->{fields}->{sql};
	my $res = $sql->previousSHAPasswords($anr);
	my $count = scalar(@$res);
	$count;
}

sub fetchCurrentRcryptPasswordViaSQL {
	my $self = shift;
	my $anr = $self->{fields}->{anr};
	my $sql = $self->{fields}->{sql};
	my $pwd = $sql->getCurrentRcryptPassword($anr);
	$pwd =~ s/^\{rcrypt\}// if $pwd;
	return $pwd;
}


sub fetchCurrentSQLPassword {
	my $self = shift;
	my $anr = $self->{fields}->{anr};
	my $rcryptPwd = $self->fetchCurrentRcryptPasswordViaSQL();
	if ($rcryptPwd) {
		return Rcrypt::decrypt($rcryptPwd, $self->{rcryptkey});
	} else {
		return undef;
	}
}

sub fetchCurrentLdapPassword {
	# Fetch the current password from the 'ldaphost' ldapobject
	my $self = shift;
	my $ldaphost = $self->{fields}->{ldaphost};
	my $anr = $self->{fields}->{anr};
	$ldaphost =~ s{ldaps?://}{};

	$self->debug("ldaphost: $ldaphost");
	my $ldapobject = $self->{ldapobjects}->{$ldaphost};

	my $entries = $ldapobject->{entries};
	my $entry = @$entries[0];
	die "No entry found for ldaphost: $ldaphost" unless $entry;

	my $rcryptpwd =	$entry->get_value('rcryptpassword');
	unless ($rcryptpwd) {
		warn "NO RCRYPTPWD FOUND for: $self->{fields}->{anr}";
		return (undef, undef);
	}
	my $orgRcrypt = $rcryptpwd;
	die "invalid rcrypt found" unless $rcryptpwd=~s/\{.*?\}//;
	my $decrypted = Rcrypt::decrypt($rcryptpwd, $self->{rcryptkey});
	return ($orgRcrypt, $decrypted);
}

sub synchronisePassword {
	my $self = shift;
	my $ret;
	$self->{fields}->{userinfo}->{suppressStatusUpdate} = 1;
	$self->{fields}->{allowpwdrepeat} = 1;
	$self->{fields}->{skiporacle} = 1 ;#if $self->{fields}->{passwordLeader} eq 'oracle';
	$self->{fields}->{synchronisePassword} = 1;

	unless ( grep /^$self->{fields}->{passwordLeader}$/, qw (oracle ldap)){
		die "$FAIL incorrect passwordleader: $self->{fields}->{passwordLeader}";
	}

	#
	# fixme is dat passwordLeader idee nog wel nodig, 
	# of mag het tegenwoordig gewoon altijd 'ldap' zijn?
	#
	my $passwordLeader = $self->{fields}->{passwordLeader};

	$ret->{oracle} = $self->fetchCurrentSQLPassword();
	($ret->{rcrypt}, $ret->{ldap}) = $self->fetchCurrentLdapPassword();
	$self->{fields}->{newPwd} = $ret->{$passwordLeader};

	if($ret->{rcrypt} and $ret->{oracle} and $ret->{ldap}){
		if ($ret->{oracle} eq $ret->{ldap}) {
			$self->debug("No need to sync according to baseldapserver and Oracle, syncing anyway");
		}
	}

	unless ($self->{fields}->{newPwd}){
		$self->debug("no password yet");
		if ($ret->{ldap}) {
			$self->{fields}->{newPwd} = $ret->{ldap}
		} else {
			die("Error: No password found for: $self->{fields}->{anr}");
		}
	}
	$self->setPassword();
}


sub getparms {
	my $self = shift;
	my $cf = shift;
	my $h;
	my $cfhash = $cf->hash();

	foreach my $el (keys %$cfhash) {
		$h->{$el} = $cfhash->{$el};
	}

	my $userinfo = $self->{fields}->{userinfo};
	foreach my $el (keys %$userinfo) {
		$h->{$el} = $userinfo->{$el};
	}
	$h->{requestfrom} = join ("\n", map {s/\n//; $_;} $cf->value('requestfrom'))
		if $cfhash->{requestfrom};

	$h->{'ldapservers'} = [($cf->value('ldapservers'))];
	$self->error("Config error ",$cf->error()) if ($cf->error());

	#optional untouchables, silent clients etc
	my $untouchables = [($cf->value('untouchables'))];

	$cf->clearerror() if $cf->error();

	if ($untouchables->[0])	{
		$h->{'untouchables'} = $untouchables;
		$self->debug( "untouchables: \"@{$h->{untouchables}}\"" ) ;
	}

	my $silentsshclients = [($cf->value('silentsshclients'))];
	$cf->clearerror() if $cf->error();

	my $silentclients = [($cf->value('silentclients'))];
	if ($silentclients->[0]) {
		$h->{silentclients} = $silentclients ;
		$self->debug("Silentclients: \"@{$h->{silentclients}}\"") ;
	}

	return $h;
}

sub getAdminAccounts {
	my $self = shift;
	my $secretfile ="$main::SYSCONFDIR/$self->{fields}->{nosvnfile}";;

	my $secretcf = new Readconfig (
		{ 'configfile' => $secretfile,
		 'needs' => [qw (ldapaccounts rcryptkey oracleaccount)],
		 'allows' => [ '.*' ],
		 'secret'=>1
		});

	if ($secretcf->error()) {
		$self->error($secretcf->error()) ;
		return;
	}

	$self->{secretcf} = $secretcf;

	# rcryptkeys
	$self->{rcryptkey} = $secretcf->value('rcryptkey');

	# privileged ldap accounts
	foreach my $hostspec ($secretcf->value('ldapaccounts')){
		next unless $hostspec;
		my ($host, $account, $pwd) = split(/\s*;\s*/, $hostspec, 3);
		$self->{ldapaccounts}->{$host}->{account} = $account;
		$self->{ldapaccounts}->{$host}->{pwd} = $pwd;
	}
	my $oracle;
	($oracle->{host}, $oracle->{account},$oracle->{pwd}) = split(/\s*;\s*/, $secretcf->value('oracleaccount'));

	die $secretcf->error() if $secretcf->error();

	$self->{oracleaccount} = $oracle;
	#combine this with the other configfile, represented in fields:

	# restrict possible subkeys, create hash from list
	my $subkeys;
	map {$subkeys->{$_} = 1;} qw (host attribs base testaccount onusernotfound hashtype deletetype
			    uid limit2 onserverdown ldapidentifier additionalfilter bindmethod
	            passwordfield forcedrepeats bindverification);

	my $servers;

	#
	# use some unique delimitor
	#
	my $delim = '{}';
	my $delimexpr = $delim; $delimexpr =~ s/(\W)/\\$1/g;

	my $ldapservers= join ($delim, @{$self->{fields}->{ldapservers}});
# warn "ldapservers: $ldapservers";
	foreach my $xhost (split /\s*;\s*/ , $ldapservers) {
		my $h;
		foreach my $keypair (split /\s*$delimexpr\s*/ , $xhost) {
			next unless ($keypair);
			my ($key,$value) = split(/\s*=\s*/, $keypair, 2);
			if (exists($subkeys->{$key})) {
				$h->{$key} = $value;
			} else {
				warn "Error: invalid subkey: \"$key\" of pair: \"$keypair\" in: $xhost\nin file: keymaster.cf";
			}
		}
		$servers->{$h->{host}}=$h;
		delete($h->{host});
	}

	foreach my $xhost (keys %$servers) {
		my $port = 389;
		my $protocol;
		my $host = $xhost;
		$host = "ldap://$host" unless $host =~ /^ldaps?/i;
		($protocol, $host)=split( /:\/\//, $host);
		($host, $port) = split(/:/,$host) if $host =~ /:/;
		$self->debug("protocol: $protocol, host: $host, port $port");

		if (exists ($self->{ldapaccounts}->{$host})) {
			# merge two hashes
			map { $self->{ldapaccounts}->{$host}->{$_} = $servers->{$xhost}->{$_}} (keys (%{$servers->{$xhost}}));
			$self->{ldapaccounts}->{$host}->{port} = $port;
			$self->{ldapaccounts}->{$host}->{protocol} = $protocol;
		} else {
			my $panic = "NO LDAPACCOUNT SPECIFIED in file \"$secretfile\" for host: \"$xhost\" (... spelled correctly?)";
			warn $panic;
			$self->debug($self->inspect($self->{ldapaccounts}));
			confess "configError, $panic";
		}
	}

	foreach my $host (keys (%{$self->{ldapaccounts}})) {
		if (!exists($self->{fields}->{userinfo}->{studentaccount})) {
			my $limit2 = $self->{ldapaccounts}->{$host}->{limit2};
			if (defined $limit2 && $limit2 eq 'studentaccounts') {
				delete $self->{ldapaccounts}->{$host};
				$self->debug("skipping $host, because $self->{fields}->{userinfo}->{uid} is no studentaccount");
				next ;
			}
		}

		unless ($self->{ldapaccounts}->{$host}->{protocol}) {
			warn "Warning: host: $host not defined in keymaster.cf";
		}
	}
}

sub genPassword {
	my $self = shift;
	my $size = 15;
	my $string;
	open FH, '/dev/urandom' or die "Couldn't open /dev/urandum!";
	while ($size) {
		read FH, $_, 1;
		next unless /[A-Za-z0-9]/;
		$string .= $_;
		$size--;
	}
	close FH;
	return $string;
}

sub generateRandomPassword {
	my $self = shift;
	my ($capital, $lower, $digit, $s, $punctuation);
	do {
		$s = $self->genPassword();
		$capital = $s =~ /([A-Z])/;
		$lower = $s =~ /([a-z])/;
		$digit = $s =~ /\d/;
	} while (!($capital && $lower && $digit));
	return $s;
}


sub shapePassword {
	my ($self, $server, $cleartext) = @_;
	my $password;
	my $deletetype = $self->{ldapaccounts}->{$server}->{deletetype};

	# Als het nieuwe password leeg is, gooi het attribuut weg als deletetype eq 'delete',
	# anders genereer een random password.
	if ($cleartext eq '') {
		if (defined ($deletetype) && $deletetype eq 'delete') {
			return '' ;
		} else {
			$cleartext = $self->generateRandomPassword();
		}
	}

	# allow preference for sha, default is ssha
	my $hashtype = $self->{ldapaccounts}->{$server}->{hashtype};
	if (defined $hashtype){
		if ($hashtype eq 'sha') {
			$password = $self->sha($cleartext);
		} elsif ($hashtype eq 'exop') {
			$password = $cleartext;
		} elsif ($hashtype eq 'MickeySoft') {
			$password = $self->activeDirectoryPwd($cleartext);
		} elsif ($hashtype eq 'ssha') {
			$password = $self->ssha($cleartext);
		} else {
			warn "undefined hashtype: '$hashtype' using default: ssha";
			$password = $self->ssha($cleartext);
		}
	} else {
		$password = $self->ssha($cleartext);
	}

	$self->{ldapaccounts}->{$server}->{verificationPassword} = $cleartext;
	return $password;
}

sub disablePassword {
	my ($self, $reason) = @_;
	warn "DisabledReason: $reason" if $reason;
	$self->{fields}->{userinfo}->{suppressStatusUpdate} = 1;
	$self->{fields}->{newPwd} = '';
	$self->{fields}->{skiporacle} = 1;
	$self->{fields}->{allowpwdrepeat} = 1;
	$self->{fields}->{disableReason} = $reason if $reason;
	my $passwordStatus = $self->{fields}->{userinfo}->{passwordStatus};
	my $uid = $self->{fields}->{userinfo}->{uid};

	warn "Disabling password for $uid";
	$self->setPassword();
}


sub setPassword {
	my $self = shift;
	# bewaar huidige password gegevens
	my ($curPwdRcrypted, $curPwdClearText) = $self->fetchCurrentLdapPassword();

	# dit kan vast wel met een hash
	if (exists(($self->{fields}->{untouchables}))) {
		if (grep(/$self->{fields}->{anr}/, @{$self->{fields}->{untouchables}})) {
			warn "$FAIL $self->{fields}->{anr} is an UNTOUCHABLE! The password remains untouched :) \n";
			exit 0;
		}
	}

	$self->{fields}->{dryrun} = 'no' unless exists ($self->{fields}->{dryrun});
	$self->debug("Dryrun state: " . $self->{fields}->{dryrun});
	my $newfields;
	my $newPwd = $self->{fields}->{newPwd} // '';
	my $userinfo = $self->{fields}->{userinfo};
	my $suppressStatusUpdate = $userinfo->{suppressStatusUpdate};
	$self->debug("SetPassword; suppressStatusUpdate: $suppressStatusUpdate");

	if ($newPwd){
		$newfields->{aselectaccountenabled} = 'true';
		$newfields->{passwordinvalidated} = '';
	} else {
		$newfields->{aselectaccountenabled} = 'false';
	}

	unless ($suppressStatusUpdate) { 
		my $time = time;
		$newfields->{datepasswdchanged} = $time;
		$newfields->{passwordchangedate} = strftime("%Y-%m-%d %T GMT", gmtime($time));
		my $dir = $self->{fields}->{dir};
		$dir->resetPasswordExpirationStatus($userinfo);
		if ($newPwd) {
			$newfields->{rcryptpassword} = $self->rcrypt($newPwd, $self->{rcryptkey});
			$newfields->{passwordchangecount} = $self->getPasswordChangeCount() + 1;
		}
	}

	my $type = $self->{fields}->{disableReason} // 'rcrypt';

	if ($self->{fields}->{disableReason} || $self->{fields}->{synchronisePassword}) {
		$newfields->{rcryptpassword} = $curPwdRcrypted;
	} 

	$newfields->{rcryptpassword} =~ s/^\{.*?}/{$type}/ if $newfields->{rcryptpassword};
	$newfields->{lockouttime} = 0;

	#prepare oracle changes, no commit yet
	unless ($self->{fields}->{dryrun} eq 'yes')	{
		$self->prepare_writeSQL($newfields);
		return if $self->error;
	}

	my $fail = 0;
	my $changed = {};
	my $uid = $self->{fields}->{userinfo}->{uid};

	foreach my $server (keys (%{$self->{ldapobjects}}))	{
		last if $fail;
		my $hashtype = $self->{ldapaccounts}->{$server}->{hashtype};
		my $errorString ;
		my $entries = $self->{ldapobjects}->{$server}->{entries};

		foreach my $entry (@$entries) {
			my $pwdSetAttemptsLeft = $self->{ldapaccounts}->{$server}->{forcedrepeats} || 1;
			my $dn = $entry->dn();
			my $objectclass = $entry->get_value('objectclass');
			my @objectclassList = $entry->get_value('objectclass');

			# retry loop
			do {
				$self->debug("pwdSetAttemptsLeft: $pwdSetAttemptsLeft $dn at $server");
				my $userpassword = $self->{ldapaccounts}->{$server}->{passwordfield} || 'userpassword';
				my $shapedPassword = $self->shapePassword($server, $newPwd);

				$newfields->{$userpassword} = $shapedPassword;
				my $ldap = $self->{ldapobjects}->{$server}->{object};
				my $nrChanges = 0;
				foreach my $attrib (split (/\s*,\s*/, lc($self->{ldapaccounts}->{$server}->{attribs}))) {
					$self->debug("changer: $attrib");
					unless(exists ($newfields->{$attrib})) {
						$self->debug("No new value for attrib: $attrib for server: $server");
						next;
					}

					# lege waarden zijn normaal een opdracht tot delete, maar niet voor lockouttime
					# lockouttime wordt ook 'forcedrepeats' keer gezet. Jammer dan.
					if ($newfields->{$attrib} or $attrib eq 'lockouttime') {
						if ($attrib eq 'userpassword' and $hashtype eq 'exop') {
							$self->debug("Setting exop password for $dn");
							unless ($self->{fields}->{dryrun} eq 'yes') { 
								my $mesg = $ldap->set_password(newpasswd => $newfields->{userpassword}, 
															   user => $dn);
								if ($mesg->code) {
									warn "set_password error: ", $mesg->code(), ": ", $mesg->error();
									$fail = 1;
								}
							} else {
								warn "NO set_password for $dn on $server because of dryrun";
							}

						} else {
							$self->debug("\tchange: $attrib into $newfields->{$attrib}");
							$entry->replace($attrib => $newfields->{$attrib});
							$nrChanges++;
						}
					} else {
						if ($entry->get_value($attrib)){
							$self->debug("\tdelete: $attrib");
							$entry->delete($attrib);
							$nrChanges++;
						} else {
							$self->debug("Attribute \"$attrib\" already gone for ", $entry->get_value('uid'));
						}
					}
				}
				unless ($self->{fields}->{dryrun} eq 'yes') {
					if ($nrChanges) {
						my $msg = $entry->update($ldap);
						if ($msg->code) {
							$errorString = "Ldap update ERROR: $server, $dn ($uid) " . $msg->error();
							warn $errorString;
							$fail = 1;
						} else {
							if ($self->{ldapaccounts}->{$server}->{bindverification}) {
								sleep 2;
								eval {$self->verifyPasswordSet($server)};
								$errorString = $@;
								unless ($errorString) {
									$self->verbose( "Password changed en verified on server: $server $dn ($uid)");
									$changed->{$server} = $entries;
								} else {
									warn "bind verification error: $errorString";
								}
							} else {
								$self->verbose("Password changed on server: $server $dn ($uid)");
							}
						}
					} else {
						warn "No changes needed for $dn ($uid) on server $server";
					}
				} else {
					warn "** NOT CHANGING PASSWORD ** for $dn ($uid) on server $server, because of dryrun";
				}
				
	
			} while (--$pwdSetAttemptsLeft >0);
			# Endof: retry loop

			if ($errorString) {
				$fail = 1;
				last;
			}
		} # endof foreach entry
	} #endof foreach server

#  COMMENT NEXT LINE!
#	$fail = 1;

	if ($fail){
		foreach my $server (keys (%$changed)) {
			$self->debug("RESTORING server: \u$server");
			my $userpassword = $self->{ldapaccounts}->{$server}->{passwordfield} || 'userpassword';
			my $ldap = $self->{ldapobjects}->{$server}->{object};
			my $entryList = $self->{ldapobjects}->{$server}->{entries};

			foreach my $entry (@$entryList){
				my $dn = $entry->dn();
				foreach my $attrib (split (/\s*,\s*/, lc($self->{ldapaccounts}->{$server}->{attribs}))) {
					my $value;
					if($attrib eq $userpassword){
						$value = $self->shapePassword($server, $curPwdClearText);

					} else {
						$value = $entry->get_value($attrib);
						unless (defined ($value)){
							warn "Cannot restore attrib: $attrib for $self->{fields}->{anr}";
						}
					}
					$entry->replace($attrib, $value);
				}

				my $msg = $entry->update($ldap);
				if ($msg->code) {
					warn("Ldap RESTORE error! $server, $dn ", $msg->error());
				}
				else {
					$self->debug("Ldap RESTORE succeeded for: $server, $dn");
				}
			}
		}

		$self->finalize_write2oracleSQL(0);
		die "password modification error, quiting";
	}
	else {
		$self->finalize_write2oracleSQL(1);
	}
}


sub saveldapobject {
	my $self = shift;
	my $host = shift;
#  warn $self->inspect($self->{ldapaccounts});
	$self->debug("saveldapobject host: $host");
	my $port = $self->{ldapaccounts}->{$host}->{port};

	$self->debug("port: $port, host:$host");
	my $account = $self->{ldapaccounts}->{$host}->{account};
	my $base   = $self->{ldapaccounts}->{$host}->{base};
	my $pwd   = $self->{ldapaccounts}->{$host}->{pwd};
	my $protocol = $self->{ldapaccounts}->{$host}->{protocol};

	my $ldap;

	my $retries = $self->{fields}->{ldapRetries};
	my $attempt = 0;
	my $state;
	while (!$ldap && $attempt++ < $retries) {
		$self->verbose("Connecting to ldapserver: $host, attempt: $attempt");
		$ldap = Net::LDAPS->new(
			$host,
#    FIXME: verify uitgezet voor AD
			'verify' => 'require',
			'capath' => $self->{fields}->{capath},
			'port'  => $port,
			);
		$state = $@;
	}

	unless ($ldap)	{
		warn "Could not connect to $protocol://$host:$port $@";
		if (exists ($self->{ldapaccounts}->{$host}->{onserverdown}) and
			$self->{ldapaccounts}->{$host}->{onserverdown} eq lc ('skip'))	{
			$self->debug("WARNING: Skipping host: $host, because of 'onserverdown is skip'");
			delete ($self->{ldapaccounts}->{$host});
			delete ($self->{ldapobjects}->{$host});
			return;
		} else {
			print "$FAIL Could not connect to ldapserver: $protocol://$host";
			exit 1;
		}
	}

	my $bindmethod = $self->{ldapaccounts}->{$host}->{bindmethod};
	my ($dn, $msg, $entry);
	if (defined ($bindmethod && $bindmethod eq 'dn')){
		$dn = $account;

	} else {
		my $accountbase=$base;
		if ($account=~ /,/) {
			($account,$accountbase) = split (/\s*,\s*/ , $account, 2);
		}

		# search for the special account
		$msg = $ldap->search(base => $accountbase, filter => $account);
		if ($msg->code)	{
			warn ("Error using filter: $account in ldapserver: $protocol://$host, accountbase: $accountbase, ", $msg->error);
			exit 1;
		}

		$entry = ($msg->entries())[0];
		unless (defined ($entry)) {
			warn("User \"$account\" unknown on ldapserver: $protocol://$host, accountbase: $accountbase");
			exit 1;
		}
		$dn = $entry->dn();
	}

	$msg = $ldap->bind($dn, password=>$pwd);
	if ($msg->code)	{
		die "error binding $dn on ldapserver: $protocol://$host:$port ",$msg->error();
	}

	$self->debug("successful bind to ldap server: $host, binder:$dn, port: $port, base:$base");
	$self->{ldapobjects}->{$host}->{object}=$ldap;
	my $filter = "employeenumber=$self->{fields}->{anr}";

	# FIXME: gedrag controleren
	if (exists(($self->{ldapaccounts}->{$host}->{limit2}))){
		return if
			($self->{ldapaccounts}->{$host}->{limit2} eq 'students') &&
			(! $self->{fields}->{userinfo}->{studentaccount} );
	}

	my $uid = $self->{fields}->{userinfo}->{uid};
	if (exists $self->{ldapaccounts}->{$host}->{ldapidentifier}){
		my $saved = $filter;
		my $identifier = $self->{ldapaccounts}->{$host}->{ldapidentifier};
		$filter = "$identifier=$uid";
		$self->verbose("ldapidentifier CHANGING FILTER from '$saved' with '$filter' for host: $host");
	}

	if (exists ($self->{ldapaccounts}->{$host}->{testaccount})){
		$self->debug("Overruling account: $filter with $self->{ldapaccounts}->{$host}->{testaccount} for host: $host");
		$filter = $self->{ldapaccounts}->{$host}->{testaccount};
	}

	$msg = $ldap->search(
		base => $base,
		filter => "(&($filter)(objectclass=person))"
		);
	if ($msg->code)	{
		warn "error using filter: $filter in ldapserver: $host", $msg->error();
		$entry = undef;
	} else {
		$entry=($msg->entries())[0];
	}

	unless (defined($entry)) {
#    warn $self->inspect($self->{ldapaccounts}->{$host});
		if ( exists ($self->{ldapaccounts}->{$host}->{onusernotfound})) {
			my $skip = 0;
			if ($self->{ldapaccounts}->{$host}->{onusernotfound} eq 'skip') {
				$self->verbose( "SKIPPING server: $host, since user: \"$filter\" was not found");
				$skip = 1;
			} elsif ($self->{ldapaccounts}->{$host}->{onusernotfound} eq 'skipunlessstudent') {
				unless (exists($self->{fields}->{userinfo}->{studentaccount})) {
					$skip = 1;
					$self->verbose("SKIPPING server: \U$host, since user: \"$filter\" was not found plus user is not student");
					$self->debug($self->inspect($self->{fields}->{userinfo}));
				}
			}

			if ($skip) {
				delete($self->{ldapaccounts}->{$host});
				delete($self->{ldapobjects}->{$host});
				return;
			}
		} else {
			warn("Nor results for filter: \"$filter\" on ldapserver: $host ");
			print("$FAIL ldap_unknown_user, $filter\n");
			exit 0;
		}
	} else {
		$self->verbose( "Found user: $filter, on: $host");
		# save the current attribs for restoring purposes for this user
		$dn = $entry->dn();
		$self->{ldapobjects}->{$host}->{dn} = $dn;
		$self->{ldapobjects}->{$host}->{entries} = [$msg->entries()];
	}
}

sub checkAccountInSQL {
	my $self = shift;
	my $anr = $self->{fields}->{anr};
	my $sql = $self->{fields}->{sql};
	my $oUid = $sql->getUsername($anr);

	unless (defined ($oUid)){
		$self->error ("userError: No valid account in mibase for anr: '$anr'");
	} 
	return $oUid;
}


sub update_passwordSQL {
	my ($self, $anr, $newfields) = @_;
	my $newPwd = $self->{fields}->{newPwd};

	my $pw_rcrypt_after = $newfields->{rcryptpassword};
	my $pw_sha_after = $self->ssha($newPwd);
	chomp ($pw_sha_after);

	my $fullldaphost = $self->{fields}->{ldaphost};
	my ($ldaphost) = $fullldaphost=~/^[^\/]+\/\/([^:]+)/;

	my $pw_rcrypt_before = $self->fetchCurrentSQLPassword() || '';
	$pw_rcrypt_before = $self->rcrypt($pw_rcrypt_before, $self->{rcryptkey});

	my $pw_sha_before = $self->fetchCurrentSHAPassword() || $self->sha('');
	my $changed_by = $self->{fields}->{userinfo}->{changed_by};
	my $uid = $self->{fields}->{userinfo}->{uid};
	my $epoch_time = $newfields->{datepasswdchanged};
	my $sql = $self->{fields}->{sql};

	$sql->setPwd($anr, $uid, $pw_sha_before, $pw_rcrypt_before,
				 $pw_sha_after, $pw_rcrypt_after, $changed_by, $epoch_time);
}


sub prepare_writeSQL {
	my $self = shift;
	my $newfields = shift;
	if ($self->{fields}->{skiporacle}) {
		$self->debug("SKIPPING ORACLE on request");
		$self->{skiporacle} = 1;
		return;
	}

	my $anr=$self->{fields}->{anr};
	if (exists ($self->{fields}->{oracle_testanr})){
		$anr = $self->{fields}->{oracle_testanr};
		$self->debug("Overwriting ORACLE anr with '$anr'");
	}

	unless ($self->{fields}->{allowpwdrepeat}) {
		if ($self->usedPasswordBefore()) {
			$self->error("userError: usedpasswordbefore");
			return;
		}
	}

  $self->update_passwordSQL($anr, $newfields);
}

sub finalize_write2oracleSQL {
	#state ==1 commit, state==0 rollback
	my ($self, $state)=@_;

	my $sql = $self->{fields}->{sql};
	my $anr = $self->{fields}->{anr};
	my $status=0;

	unless ($self->{skiporacle}) {
		if($state) {
			$sql->commit;
		} else {
			$sql->rollback;
		}
	} else {
		$self->debug("Not commiting because of 'skiporacle'");
	}
}

sub usedPasswordBefore {
	my ($self) = @_;
	my $anr = $self->{fields}->{anr};
	my $sql = $self->{fields}->{sql};
	my $newpwd = $self->{fields}->{newPwd};

	my ($curPwdRcrypted, $curPwdClearText) = $self->fetchCurrentLdapPassword();

	if ($curPwdClearText eq $newpwd){
		return 1;
	}
	
	# N.B. previousPasswords returns an array of refs to arrays
	my $a = $sql->previousPasswords($anr);
	foreach my $encr (@$a) {
		return 1
			if $newpwd eq Rcrypt::decrypt($encr, $self->{rcryptkey});
	}

   	return 0;
}

####################################################################
sub sha {
	my ($self, $cleartext) = @_;
	my $ctx = Digest::SHA->new;
	$ctx->add($cleartext);
	return '{SHA}'. encode_base64($ctx->digest);
}

sub ssha {
	my ($self, $cleartext) = @_;
	my $ctx = Digest::SHA->new;
	$ctx->add($cleartext);
	my $salt = $self->genrandom(8);
	$salt = encode_base64($salt); chop($salt); chop($salt);
	$ctx->add($salt);
	return '{SSHA}'. encode_base64($ctx->digest. $salt);
}

sub activeDirectoryPwd {
	my ($self, $cleartext) = @_;
	return encode('UTF-16le', '"'. $cleartext. '"' );
}

sub genrandom {
	my $self = shift;
	my $size = shift || 256;
	my $random;
	if (open FH, '/dev/urandom' ) {
		read FH, $random, $size;
		close FH;
		return $random;
	}

	my $max = $size;
	my $min = 0;
	my $maxrange = 255;

	for (my $i = 0; $i < $max; $i++) {
		$random .= int (rand ($maxrange)) + $min;
	}
	$random;
}


sub rcrypt {
	my ($self, $cleartext, $key) = @_;
	return "{rcrypt}". Rcrypt::encrypt($cleartext, $key);
}

42;
